# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

"""Thermodynamics quantities. Values from Wikipedia unless specified."""

import numpy as np
from thermohl import floatArrayLike

# Standard temperature and pressure from EPA and NIST; _T0 in K, _p0 in Pa
_T0, _p0 = 293.15, 1.01325e05

# Boltzman constant, Avogadro number and Gas constant (all in SI)
_kb = 1.380649e-23
_Na = 6.02214076e23
_R = _kb * _Na


class Air:
    @staticmethod
    def heat_capacity(T: floatArrayLike = _T0) -> floatArrayLike:
        """In J.kg**-1.K**-1"""
        return np.interp(T, [240.0, 600.0], [1.006, 1.051])


class Water:
    @staticmethod
    def boiling_point(p: floatArrayLike = _p0) -> floatArrayLike:
        """Using Clausius–Clapeyron equation; in K."""
        # convert H from J.kg**-1 to J.mol**-1 using molar mass
        H = Water.heat_of_vaporization() * 0.018015
        Tb = 1.0 / (1 / 373.15 - _R * np.log(p / _p0) / H)
        return Tb

    @staticmethod
    def heat_capacity(T: floatArrayLike = _T0) -> floatArrayLike:
        """From NIST webbook; in J.kg**-1.K**-1.
        See https://webbook.nist.gov/cgi/cbook.cgi?Name=Water&Units=SI.
        """
        A = -203.6060
        B = +1523.290
        C = -3196.413
        D = +2474.455
        E = 3.855326
        t = T / 1000.0
        return A + B * t + C * t**2 + D * t**3 + E / t**2

    @staticmethod
    def heat_of_vaporization() -> float:
        # """At T=373.15 K and normal pressure; in J.mol**-1."""
        # return 4.0660E+04
        """At T=373.15 K and normal pressure; in J.kg**-1."""
        return 2.257e05

    @staticmethod
    def vapor_pressure(T: floatArrayLike = _T0) -> floatArrayLike:
        """Using Buck equation; in Pa."""
        Tc = T - 273.15
        return 611.21 * np.exp((18.678 - Tc / 234.5) * (Tc / (257.14 + Tc)))

    @staticmethod
    def volumic_mass(T: floatArrayLike = _T0) -> floatArrayLike:
        """In kg.m**-3."""
        xp = np.array(
            [
                0.0,
                1.0,
                2.0,
                3.0,
                4.0,
                5.0,
                6.0,
                7.0,
                8.0,
                9.0,
                10.0,
                11.0,
                12.0,
                13.0,
                14.0,
                15.0,
                16.0,
                17.0,
                18.0,
                19.0,
                20.0,
                21.0,
                22.0,
                23.0,
                24.0,
                25.0,
                26.0,
                27.0,
                28.0,
                29.0,
                30.0,
                31.0,
                32.0,
                33.0,
                34.0,
                35.0,
                36.0,
                37.0,
                38.0,
                39.0,
                40.0,
                45.0,
                50.0,
                55.0,
                60.0,
                65.0,
                70.0,
                75.0,
                80.0,
                85.0,
                90.0,
                95.0,
                100.0,
            ]
        )
        fp = np.array(
            [
                999.840,
                999.899,
                999.940,
                999.964,
                999.972,
                999.964,
                999.940,
                999.902,
                999.848,
                999.781,
                999.700,
                999.6,
                999.5,
                999.38,
                999.24,
                999.1,
                998.94,
                998.77,
                998.59,
                998.4,
                998.2,
                997.99,
                997.77,
                997.54,
                997.29,
                997.04,
                996.78,
                996.51,
                996.23,
                995.94,
                995.64,
                995.34,
                995.02,
                994.7,
                994.37,
                994.03,
                993.68,
                993.32,
                992.96,
                992.59,
                992.21,
                990.21,
                988.03,
                985.69,
                983.19,
                980.55,
                977.76,
                974.84,
                971.79,
                968.61,
                965.3,
                961.88,
                958.35,
            ]
        )
        return np.interp(T, xp, fp)


class Ice:
    @staticmethod
    def heat_capacity() -> float:
        """In J.kg**-1.K**-1"""
        return 2.093e03

    @staticmethod
    def heat_of_fusion() -> float:
        """At T=273.15 K and normal pressure; in J.kg**-1."""
        return 3.3355e05



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any


from thermohl import floatArrayLike
from thermohl.power.convective_cooling import ConvectiveCoolingBase
from thermohl.power.ieee import Air


class ConvectiveCooling(ConvectiveCoolingBase):
    """Convective cooling term."""

    def __init__(
        self,
        alt: floatArrayLike,
        azm: floatArrayLike,
        Ta: floatArrayLike,
        ws: floatArrayLike,
        wa: floatArrayLike,
        D: floatArrayLike,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Parameters
        ----------
        alt : float or np.ndarray
            Altitude.
        azm : float or np.ndarray
            Azimuth.
        Ta : float or np.ndarray
            Ambient temperature.
        ws : float or np.ndarray
            Wind speed.
        wa : float or np.ndarray
            Wind angle regarding north.
        D : float or np.ndarray
            External diameter.

        """
        super().__init__(
            alt,
            azm,
            Ta,
            ws,
            wa,
            D,
            Air.volumic_mass,
            Air.dynamic_viscosity,
            Air.thermal_conductivity,
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numbers
from typing import Dict, Any, Optional

import numpy as np
import pandas as pd

from thermohl import floatArrayLike, floatArray
from thermohl.solver.base import Solver as Solver_
from thermohl.solver.base import _DEFPARAM as DP
from thermohl.solver.base import _set_dates, reshape
from thermohl.utils import bisect_v


class Solver1T(Solver_):
    def steady_temperature(
        self,
        Tmin: float = DP.tmin,
        Tmax: float = DP.tmax,
        tol: float = DP.tol,
        maxiter: int = DP.maxiter,
        return_err: bool = False,
        return_power: bool = True,
    ) -> pd.DataFrame:
        """
        Compute steady-state temperature.

        Parameters
        ----------
        Tmin : float, optional
            Lower bound for temperature.
        Tmax : float, optional
            Upper bound for temperature.
        tol : float, optional
            Tolerance for temperature error.
        maxiter : int, optional
            Max number of iteration.
        return_err : bool, optional
            Return final error on temperature to check convergence. The default is False.
        return_power : bool, optional
            Return power term values. The default is True.

        Returns
        -------
        pandas.DataFrame
            A DataFrame with temperature and other results (depending on inputs)
            in the columns.

        """

        # solve with bisection
        T, err = bisect_v(
            lambda x: -self.balance(x), Tmin, Tmax, (self.args.max_len(),), tol, maxiter
        )

        # format output
        df = pd.DataFrame(data=T, columns=[Solver_.Names.temp])

        if return_err:
            df[Solver_.Names.err] = err

        if return_power:
            df[Solver_.Names.pjle] = self.jh.value(T)
            df[Solver_.Names.psol] = self.sh.value(T)
            df[Solver_.Names.pcnv] = self.cc.value(T)
            df[Solver_.Names.prad] = self.rc.value(T)
            df[Solver_.Names.ppre] = self.pc.value(T)

        return df

    def transient_temperature(
        self,
        time: floatArray = np.array([]),
        T0: Optional[float] = None,
        return_power: bool = False,
    ) -> Dict[str, Any]:
        """
        Compute transient-state temperature.

        Parameters
        ----------
        time : numpy.ndarray
            A 1D array with times (in seconds) when the temperature needs to be
            computed. The array must contain increasing values (undefined
            behaviour otherwise).
        T0 : float
            Initial temperature. If set to None, the ambient temperature from
            internal dict will be used. The default is None.
        return_power : bool, optional
            Return power term values. The default is False.

        Returns : Dict[str, Any]
            A dictionary with temperature and other results (depending on inputs)
            in the keys.
        """

        # get sizes (n for input dict entries, N for time)
        n = self.args.max_len()
        N = len(time)
        if N < 2:
            raise ValueError("The length of the time array must be at least 2.")

        # get initial temperature
        if T0 is None:
            T0 = (
                self.args.Ta
                if isinstance(self.args.Ta, numbers.Number)
                else self.args.Ta[0]
            )

        # get month, day and hours
        month, day, hour = _set_dates(
            self.args.month, self.args.day, self.args.hour, time, n
        )

        # Two dicts, one (dc) with static quantities (with all elements of size n), the other (de)
        # with time-changing quantities (with all elements of
        # size N*n); uk is a list of keys that are in dc but not in de.
        de = dict(
            month=month,
            day=day,
            hour=hour,
            I=reshape(self.args.I, N, n),
            Ta=reshape(self.args.Ta, N, n),
            wa=reshape(self.args.wa, N, n),
            ws=reshape(self.args.ws, N, n),
            Pa=reshape(self.args.Pa, N, n),
            rh=reshape(self.args.rh, N, n),
            pr=reshape(self.args.pr, N, n),
        )
        del (month, day, hour)

        # shortcuts for time-loop
        imc = 1.0 / (self.args.m * self.args.c)

        # init
        T = np.zeros((N, n))
        T[0, :] = T0

        # main time loop
        for i in range(1, len(time)):
            for k, v in de.items():
                self.args[k] = v[i, :]
            self.update()
            T[i, :] = (
                T[i - 1, :] + (time[i] - time[i - 1]) * self.balance(T[i - 1, :]) * imc
            )

        # save results
        dr = dict(time=time, T=T)

        # manage return dict 2 : powers
        if return_power:
            for c in Solver_.Names.powers():
                dr[c] = np.zeros_like(T)
            for i in range(N):
                for k in de.keys():
                    self.args[k] = de[k][i, :]
                self.update()
                dr[Solver_.Names.pjle][i, :] = self.jh.value(T[i, :])
                dr[Solver_.Names.psol][i, :] = self.sh.value(T[i, :])
                dr[Solver_.Names.pcnv][i, :] = self.cc.value(T[i, :])
                dr[Solver_.Names.prad][i, :] = self.rc.value(T[i, :])
                dr[Solver_.Names.ppre][i, :] = self.pc.value(T[i, :])

        # squeeze return values if n is 1
        if n == 1:
            keys = list(dr.keys())
            keys.remove(Solver_.Names.time)
            for k in keys:
                dr[k] = dr[k][:, 0]

        return dr

    def steady_intensity(
        self,
        T: floatArrayLike = np.array([]),
        Imin: float = DP.imin,
        Imax: float = DP.imax,
        tol: float = DP.tol,
        maxiter: int = DP.maxiter,
        return_err: bool = False,
        return_power: bool = True,
    ) -> pd.DataFrame:
        """Compute steady-state max intensity.

        Compute the maximum intensity that can be run in a conductor without
        exceeding the temperature given in argument.

        Parameters
        ----------
        T : float or numpy.ndarray
            Maximum temperature.
        Imin : float, optional
            Lower bound for intensity. The default is 0.
        Imax : float, optional
            Upper bound for intensity. The default is 9999.
        tol : float, optional
            Tolerance for temperature error. The default is 1.0E-06.
        maxiter : int, optional
            Max number of iteration. The default is 64.
        return_err : bool, optional
            Return final error on intensity to check convergence. The default is False.
        return_power : bool, optional
            Return power term values. The default is True.

        Returns
        -------
        pandas.DataFrame
            A dataframe with maximum intensity and other results (depending on inputs)
            in the columns.

        """

        # save transit in arg
        transit = self.args.I

        # solve with bisection
        shape = (self.args.max_len(),)
        T_ = T * np.ones(shape)
        jh = (
            self.cc.value(T_)
            + self.rc.value(T_)
            + self.pc.value(T_)
            - self.sh.value(T_)
        )

        def fun(i: floatArray) -> floatArrayLike:
            self.args.I = i
            self.jh.__init__(**self.args.__dict__)
            return self.jh.value(T_) - jh

        A, err = bisect_v(fun, Imin, Imax, shape, tol, maxiter)

        # restore previous transit
        self.args.I = transit

        # format output
        df = pd.DataFrame(data=A, columns=[Solver_.Names.transit])

        if return_err:
            df[Solver_.Names.err] = err

        if return_power:
            df[Solver_.Names.pjle] = self.jh.value(T)
            df[Solver_.Names.psol] = self.sh.value(T)
            df[Solver_.Names.pcnv] = self.cc.value(T)
            df[Solver_.Names.prad] = self.rc.value(T)
            df[Solver_.Names.ppre] = self.pc.value(T)

        return df



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pandas as pd

from thermohl.power import rte
from thermohl.solver import Args


class ExcelSheet:
    """Object to compare power terms from rte's excel sheet version 7."""

    def __init__(self, dic):
        self.args = Args(dic)
        self.args.nbc = dic["nbc"]

    def joule_heating(self, Ta, I=None):
        if I is None:
            I = self.args["I"]
        d = self.args["d"]
        D = self.args["D"]
        Rdc = self.args["RDC20"] * (
            1.0 + self.args["kl"] * (Ta - 20.0) + self.args["kq"] * (Ta - 20.0) ** 2
        )
        z = 8 * np.pi * 50.0 * (D - d) ** 2 / ((D**2 - d**2) * 1.0e07 * Rdc)
        a = 7 * z**2 / (315 + 3 * z**2)
        b = 56 / (211 + z**2)
        beta = 1.0 - d / D
        kep = 1 + a * (1.0 - 0.5 * beta - b * beta**2)
        kem = np.where(
            (d > 0.0) & (self.args["nbc"] == 3),
            self.args["km"]
            + self.args["ki"] * I / (self.args["A"] - self.args["a"]) * 1.0e-06,
            1.0,
        )
        Rac = Rdc * kep * kem
        return Rac * self.args["I"] ** 2

    def solar_heating(self):
        csm = np.array([0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334])
        O4 = csm[self.args["month"] - 1] + self.args["day"]
        O5 = self.args["hour"]
        O6 = np.deg2rad(self.args["lat"])
        Q4 = np.deg2rad(23.46 * np.sin(np.deg2rad((284 + O4) / 365 * 360)))
        Q5 = np.deg2rad((O5 - 12) * 15)
        Q6 = np.rad2deg(
            np.arcsin(np.cos(O6) * np.cos(Q4) * np.cos(Q5) + np.sin(O6) * np.sin(Q4))
        )
        q = np.maximum(
            -42
            + 63.8 * Q6
            - 1.922 * Q6**2
            + 0.03469 * Q6**3
            - 0.000361 * Q6**4
            + 0.000001943 * Q6**5
            - 0.00000000408 * Q6**6,
            0.0,
        )
        Q7 = np.sin(Q5) / (np.sin(O6) * np.cos(Q5) - np.cos(O6) * np.tan(Q4))
        Q8 = np.deg2rad(180 + np.rad2deg(np.arctan(Q7)))
        O7 = np.pi / 2
        O2 = np.arccos(np.cos(np.deg2rad(Q6)) * np.cos(Q8 - O7))
        q *= np.sin(O2)
        return q * self.args["D"] * self.args["alpha"]

    def convective_cooling(self, Ts):
        D = self.args["D"]
        Tf = 0.5 * (Ts + self.args["Ta"])
        lm = 0.02424 + 0.00007477 * Tf - 0.000000004407 * Tf**2
        rho = (
            1.293 - 0.0001525 * self.args["alt"] + 0.00000000638 * self.args["alt"] ** 2
        ) / (1 + 0.00367 * Tf)
        mu = (0.000001458 * (Tf + 273) ** 1.5) / (Tf + 383.4)
        Re = self.args["ws"] * self.args["D"] * rho / mu
        F = np.maximum(1.01 + 1.35 * Re**0.52, 0.754 * Re**0.6)
        wa = np.deg2rad(self.args["wa"])
        K = 1.194 - np.cos(wa) + 0.194 * np.cos(2 * wa) + 0.368 * np.sin(2 * wa)
        PCn = (
            3.645
            * rho**0.5
            * D**0.75
            * np.sign(Ts - self.args["Ta"])
            * np.abs(Ts - self.args["Ta"]) ** 1.25
        )
        PCf = F * lm * K * (Ts - self.args["Ta"])
        # print(f"re={Re}, kp={K}, lam={lm}")
        # print(f"pcn={PCn}, pcf={PCf}")
        return np.maximum(PCn, PCf)

    def radiative_cooling(self, Ts):
        D = self.args["D"]
        return (
            17.8
            * D
            * self.args["epsilon"]
            * (((273 + Ts) / 100) ** 4 - ((273 + self.args["Ta"]) / 100) ** 4)
        )


def excel_conductor_data():
    """Get conductor data from excel sheet (hard-coded)."""
    df = pd.DataFrame(
        dict(
            conductor=[
                "ACSS1317",
                "Aster228",
                "Aster570",
                "Crocus412",
                "Pastel228",
                "Petunia612",
            ],
            D=[44.0, 19.6, 31.06, 26.4, 19.6, 32.1],
            d=[21.28, 0.0, 0.0, 12.0, 8.4, 13.25],
            A=[1317, 228, 570, 412, 228, 612],
            a=[0, 0, 0, 0, 0, 0],
            B=[1049, 228, 570, 323, 185, 508],
            RDC20=[0.0272, 0.146, 0.0583, 0.089, 0.18, 0.0657],
            kl=[0.004, 0.0036, 0.0036, 0.004, 0.0036, 0.0036],
            km=[1.006, 1.0, 1.0, 1.0, 1.0, 1.006],
            ki=[0.016, 0.0, 0.0, 0.0, 0.0, 0.016],
            kq=[8.0e-07, 8.0e-07, 8.0e-07, 8.0e-07, 8.0e-07, 8.0e-07],
            nbc=[3, 0, 0, 2, 2, 3],
        )
    )

    df["a"] = df["A"] - df["B"]
    df.drop(columns=["B"], inplace=True)
    df["D"] *= 1.0e-03
    df["d"] *= 1.0e-03
    df["A"] *= 1.0e-06
    df["a"] *= 1.0e-06
    df["RDC20"] *= 1.0e-03

    return df


def scenarios():
    """Get list of hard-coded scenarios to test."""
    dic = dict(
        conductor=[
            "Aster228",
            "Pastel228",
            "Petunia612",
            "Petunia612",
            "ACSS1317",
            "ACSS1317",
            "Aster228",
            "Aster228",
            "Aster228",
            "Aster228",
        ],
        Ta=[20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0],
        ws=[3.0, 3.0, 3.0, 0.0, 0.0, 3.0, 0.6, 0.6, 0.6, 0.6],
        wa=[90.0, 90.0, 90.0, 45.0, 45.0, 90.0, 90.0, 90.0, 90.0, 90.0],
        lat=[46.0, 46.0, 46.0, 46.0, 46.0, 46.0, 46.0, 46.0, 46.0, 46.0],
        alt=[1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
        azm=90.0,
        I=[1000.0, 1000.0, 1800.0, 1100.0, 3000.0, 4000.0, 700.0, 700.0, 700.0, 700.0],
        alpha=0.9,
        epsilon=0.8,
        tb=0.0,
        month=[3, 3, 3, 3, 3, 3, 3, 6, 6, 6],
        day=[7, 7, 7, 7, 7, 7, 7, 21, 21, 21],
        hour=[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 12.0, 12.0, 19.0, 12.0],
    )
    df = pd.DataFrame(dic)
    dg = excel_conductor_data()
    df = pd.merge(df, dg, on="conductor", how="left").drop(columns="conductor")
    return df


def test_compare_power():
    """Compare computed values to hard-coded ones from ieee guide [find ref]."""

    T = np.linspace(-50, +250, 999)

    ds = scenarios()
    n = len(ds)
    ds = pd.concat(len(T) * (ds,)).reset_index(drop=True)
    T = np.concatenate([n * (t,) for t in T])

    from thermohl.utils import df2dct

    d1 = df2dct(ds)
    ds["wa"] = np.rad2deg(
        np.arcsin(np.sin(np.deg2rad(np.abs(ds["azm"] - ds["wa"]) % 180.0)))
    )
    d2 = df2dct(ds)
    del (ds, n)

    pj = rte.JouleHeating(**d1)
    ps = rte.SolarHeating(**d1)
    pc = rte.ConvectiveCooling(**d1)
    pr = rte.RadiativeCooling(**d1)
    ex = ExcelSheet(d2)

    assert np.allclose(ex.joule_heating(T), pj.value(T))
    assert np.allclose(ex.solar_heating(), ps.value(0.0))
    assert np.allclose(ex.convective_cooling(T), pc.value(T))
    assert np.allclose(ex.radiative_cooling(T), pr.value(T))


def test_solar_heating():
    # adapted from devin repo dev/test/test_rte.py

    n = 5
    ones = np.ones(n)

    lat = np.array([40.0, 46.0, 46.0, 46.0, 46.0])
    azm = np.array([90.0, 0.0, 0.0, 0.0, 0.0])
    month = np.array([7, 3, 3, 3, 3])
    day = np.array([19, 7, 14, 7, 7])
    hour = np.array([14.0, 12.0, 17.0, 12.0, 12.0])
    D = 4.4e-02 * ones
    alpha = 0.9 * ones

    p = np.array([34.9, 21.9357, 13.95, 21.9357, 21.9357])
    s = rte.SolarHeating(lat, azm, month, day, hour, D, alpha, srad=None)

    assert np.allclose(p, s.value(ones), 0.1)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

"""Functions and classes to obtain various probability distributions for scalar random variables."""

import warnings
from typing import Union, Optional

import numpy as np
import scipy
from scipy.special import erf
from scipy.special import i0
from scipy.special import i1

from thermohl import floatArrayLike, frozen_dist

_twopi = 2 * np.pi


def _phi(x: float) -> float:
    """PDF of standard normal distribution."""
    return np.exp(-0.5 * x**2) / np.sqrt(_twopi)


def _psi(x: float) -> float:
    """CDF of standard normal distribution."""
    return 0.5 * (1 + erf(x / np.sqrt(2)))


def _truncnorm_header(
    a: float, b: float, mu: float, sigma: float
) -> tuple[float, float, float, float]:
    """Utility code factoring."""
    m = mu
    s = sigma
    al = (a - m) / s
    bt = (b - m) / s
    return al, bt, m, s


def _truncnorm_mean_std(
    a: float, b: float, mu: float, sigma: float
) -> tuple[float, float]:
    """Real mean and std of truncated normal distribution."""
    al, bt, m, s = _truncnorm_header(a, b, mu, sigma)
    Z = _psi(bt) - _psi(al)
    mean = m + s * (_phi(al) - _phi(bt)) / Z
    std = s * np.sqrt(
        1 + (al * _phi(al) - bt * _phi(bt)) / Z - ((_phi(al) - _phi(bt)) / Z) ** 2
    )
    return mean, std


def truncnorm(
    a: float,
    b: float,
    mu: float,
    sigma: float,
    err_mu: float = 1.0e-03,
    err_sigma: float = 1.0e-02,
    rel_err: bool = True,
) -> frozen_dist:
    """Truncated normal distribution. Wrapper from scipy.stats."""
    if a >= b:
        raise ValueError("Input a (%.3E) should be lower than b (%.3E)." % (a, b))
    if mu < a or mu > b:
        raise ValueError(
            "Input mu (%.3E) should be in [a, b] range (%.3E, %.3E)." % (mu, a, b)
        )
    if sigma < 0.0:
        raise ValueError("Input sigma (%.3E) should be positive." % (sigma,))

    al, bt, m, s = _truncnorm_header(a, b, mu, sigma)
    dist = scipy.stats.truncnorm(al, bt, m, s)

    mean = dist.mean()
    std = dist.std()

    err_mu_ = err_mu
    err_sigma_ = err_sigma
    if rel_err:
        err_mu_ *= mu
        err_sigma_ *= sigma
    if np.abs(sigma - std) >= err_sigma_:
        warnings.warn(
            "Required std cannot be achieved (%.3E instead of %.3E). Choose a lower std, extend your "
            "bounds or change your distribution." % (sigma, std),
            RuntimeWarning,
        )
    if np.abs(mu - mean) >= err_mu_:
        warnings.warn(
            "Required mean cannot be achieved (%.3E instead of %.3E). Move the mean towards the center of your "
            "bounds, extend your bounds or change your distribution." % (mu, mean),
            RuntimeWarning,
        )

    return dist


class WrappedNormal(object):
    """Wrapped-Normal distribution. Not as complete as a scipy.stat distribution."""

    def __init__(self, mu: float, sigma: float, lwrb: float = 0.0):
        if sigma < 0:
            raise ValueError("Std should be positive.")
        if mu < lwrb or mu >= lwrb + _twopi:
            raise ValueError(
                "Mean should be greater than lower bound and lower than lower bound + 2*pi."
            )
        self.lwrb = lwrb
        self.uprb = lwrb + _twopi
        self.mu = mu
        self.sigma = sigma
        return

    def rvs(
        self,
        size: int = 1,
        random_state: Optional[
            Union[int, np.random.Generator, np.random.RandomState]
        ] = None,
    ) -> floatArrayLike:
        smpl = scipy.stats.norm.rvs(
            loc=self.mu, scale=self.sigma, size=size, random_state=random_state
        )
        smpl = smpl % _twopi
        smpl[smpl < self.lwrb] += _twopi
        smpl[smpl > self.uprb] -= _twopi
        return smpl

    def mean(self) -> float:
        return self.mu

    def median(self) -> float:
        return self.mu

    def var(self) -> float:
        return 1 - np.exp(-0.5 * self.sigma**2)

    def std(self) -> float:
        return np.sqrt(self.var())

    def ppf(self, q: float) -> np.float64:
        return np.quantile(self.rvs(9999), q)


def wrapnorm(mu: float, sigma: float) -> WrappedNormal:
    """Get Wrapped Normal distribution.
    -- in radians, in [0, 2*pi]
    """
    mu2 = mu % _twopi
    if mu2 != mu:
        warnings.warn(
            "Changed mean from %.3E to %.3E to fit [0,2*pi] interval." % (mu, mu2),
            RuntimeWarning,
        )
    if sigma >= _twopi:
        warnings.warn(
            "Required std cannot be achieved (%.3E > 2*pi). Choose a lower std "
            "or change your distribution." % (sigma,),
            RuntimeWarning,
        )
    return WrappedNormal(mu2, sigma)


def _vonmises_circ_var(kappa: float) -> float:
    """Von Mises distribution circular variance."""
    return 1 - i1(kappa) / i0(kappa)


def _vonmises_kappa(sigma: float) -> float:
    """Get von Mises parameter that matches std in input."""
    from scipy.optimize import newton

    vr = 1.0 - np.exp(-0.5 * sigma**2)
    k0 = 0.5 / vr

    def fun(x: float) -> float:
        return _vonmises_circ_var(x) - vr

    try:
        kappa = newton(fun, x0=k0, tol=1.0e-06, maxiter=32)
    except RuntimeError:
        kappa = 1 / sigma**2

    return kappa


def vonmises(mu: float, sigma: float) -> frozen_dist:
    """Get von Mises distribution.
    -- in radians, in [-pi,+pi]
    """
    mu2 = mu % _twopi
    if mu2 > np.pi:
        mu2 -= _twopi
    if mu2 != mu:
        warnings.warn(
            "Changed mean from %.3E to %.3E to fit [-pi,+pi] interval." % (mu, mu2),
            RuntimeWarning,
        )
    if sigma < 0.0:
        raise ValueError("Input sigma (%.3E) should be positive." % (sigma,))
    sigmax = _twopi / np.sqrt(12)
    if sigma >= sigmax:
        warnings.warn(
            "Required std cannot be achieved (%.3E > %.3E). Choose a lower std "
            "or change your distribution." % (sigma, sigmax),
            RuntimeWarning,
        )
    kappa = _vonmises_kappa(sigma)
    return scipy.stats.vonmises_line(kappa, loc=mu)
